# ------------------------------------------------------------------------------
# WebDAV for Tclhttpd
# 
# INSTALLATION:
# * see http://www.vogel-nest.de (go to "WebDAV")
# 
# Stefan Vogel (stefan at vogel-nest dot de)
#
# HISTORY:
# 2004-07-29: Version 0.1a
#  * initial version
# ------------------------------------------------------------------------------
package require tdom
package require uri
# needed for "sets"
package require struct
package require fileutil

# modify standard tclhttpd to match needs for webdav
if {![info exists Httpd_Errors(207)]} {
	# repair 200 Data follows (from tclhttpd 3.5.1)
	# should be OK
	# repair 403 {Permission denied} should be Forbidden
	array set Httpd_Errors {
		200 OK
		201 Created
		207 Multi-Status
		403 Forbidden
		409 Conflict
		412 "Precondition Failed"
		423 Locked
		507 "Insufficient Storage"
	}
}

# get the latest fix if not already installed ;-)
# Thanks to Stefan Finzel
if {[info procs Httpd_AppendHeader] == ""} {
	# Httpd_AppendHeader
	# Append header entries to be used in a reply
	# Call this before using Httpd_ReturnFile or
	# Httpd_ReturnData
	#
	# Arguments:
	# sock handle on the connection
	# lines append-header lines, overwrite any preexisting
	
	proc Httpd_AppendHeader {sock lines} {
		upvar #0 Httpd${sock} data
		set data(header,append,${sock}) ${lines}
	}
}


# for TclHttpd-Releases pre 3.5
if {[info procs Httpd_ReplaceHeader] == ""} {
	proc Httpd_ReplaceHeader {sock lines} {
		upvar #0 Httpd${sock} data
		set data(header,replace,${sock}) ${lines}
	}

	proc Httpd_ReturnData {sock type content {code 200} {close 0}} {
		global Httpd
		upvar #0 Httpd$sock data
		
		if {[Thread_Respond $sock [list Httpd_ReturnData $sock $type $content $code $close]]} {
			return
		}

		Count urlreply
		if {$close == 0} {
			set close [HttpdCloseP $sock]
		}
		if {[catch {
			####### (SV) change begins here #######
			if {[info exists data(header,replace,${sock})]} {
				puts ${sock} $data(header,replace,${sock})
				unset data(header,replace,${sock})
			} else {
				HttpdRespondHeader $sock $type $close [string length $content] $code
				if {[info exist data(header,append,${sock})]} {
					puts ${sock} $data(header,append,${sock})
				unset data(header,append,${sock})
				}
			}
			####### (SV) change ends here #######
			
			HttpdSetCookie $sock
			puts $sock ""
			if {$data(proto) != "HEAD"} {
				fconfigure $sock -translation binary -blocking $Httpd(sockblock)
				puts -nonewline $sock $content
			}
			Httpd_SockClose $sock $close
		} err]} {
			HttpdCloseFinal $sock $err
		}
	}

}

# we won't support asynchron file-delivery (because it doesn't work
# with vfs-mounted files)
# copy&paste of httpd.tcl/Httpd_ReturnFile
proc WebDav_ReturnFile {sock type path {offset 0}} {
    global Httpd
    upvar #0 Httpd$sock data

    if {[Thread_Respond $sock \
	    [list WebDav_ReturnFile $sock $type $path $offset]]} {
	return
    }

    # Set file size early so it gets into all log records

    set data(file_size) [file size $path]
    set data(code) 200

    Count urlreply
    if {[info exists data(mime,if-modified-since)]} {
        # No need for complicated date comparison, if they're identical then 304.
	if {$data(mime,if-modified-since) == [HttpdDate [file mtime $path]]} {
            Httpd_NotModified $sock
            return
        }
    } 

    # Some files have a duality, when the client sees X bytes but the
    # file is really X + n bytes (the first n bytes reserved for server
    # side accounting information.

    incr data(file_size) -$offset

    if {[catch {
	set close [HttpdCloseP $sock]
	HttpdRespondHeader $sock $type $close $data(file_size) 200
	HttpdSetCookie $sock
	puts $sock "Last-Modified: [HttpdDate [file mtime $path]]"
	puts $sock ""
	if {$data(proto) != "HEAD"} {
	    set in [open $path]		;# checking should already be done
	    fconfigure $in -translation binary -blocking 1
	    if {$offset != 0} {
		seek $in $offset
	    }
	    fconfigure $sock -translation binary -blocking $Httpd(sockblock)
	    set data(infile) $in
	    Httpd_Suspend $sock 0
	    # (SV) use synchron delivery
		# fcopy $in $sock -command [list HttpdCopyDone $in $sock $close]
		fcopy $in $sock
		if {$error == ""} {
			set error Close
		}
		Httpd_SockClose $sock $close $error
		# /(SV)
	} else {
	    Httpd_SockClose $sock $close
	}
    } err]} {
	HttpdCloseFinal $sock $err
    }
}

################################################################################

namespace eval de::vogel::webdav {
	namespace export webdav_resource

	# hold the various configuration of derived namespaces
	variable aConfig
	
	# to enable debugging: 
	set aConfig(debug) 1

	variable lPossibleMethods
	set lPossibleMethods [list \
							  OPTIONS \
							  PROPFIND \
							  GET \
							  PUT \
							  POST \
							  DELETE \
							  MKCOL \
							  PROPPATCH \
							  HEAD \
							  TRACE \
							  COPY \
							  MOVE \
							  LOCK \
							  UNLOCK \
							  ]

	# lAllowedMethods are determined by scan over module-functions
}

proc de::vogel::webdav::log {msg} {
	variable aConfig
	if {[info exists aConfig(debug)]
		&& $aConfig(debug)} {
		puts stderr $msg
	}
}

proc de::vogel::webdav::getMethod {prefix method} {
	variable aConfig
	return [namespace inscope :: "info procs $aConfig($prefix,namespace)::$method"]
}

proc de::vogel::webdav::readConfig {{filename webdav_config.tcl}} {
	variable aConfig
	array set aConfig {}
	source $filename
}

proc de::vogel::webdav::webdav_resource {path argList} {
	variable aConfig
	foreach {key value} $argList {
		set aConfig($path,$key) $value
	}
	# install the Direct_URL
	Url_PrefixInstall $path [list de::vogel::webdav::handler $path]
	uplevel #0 "source $aConfig($path,filename)"
	log "WebDAV-Url on $path"
}

proc de::vogel::webdav::getResource {path key} {
	variable aConfig
	return $aConfig($path,$key)
}

proc de::vogel::webdav::path2RealPath {urlPath} {
	global Config
	return [file join $Config(docRoot) [string trimleft $urlPath /]]
}

proc de::vogel::webdav::realPath2UrlPath {realPath} {
	global Config
	return "/[::fileutil::stripN $realPath [llength [file split $Config(docRoot)]]]"
	
}

proc de::vogel::webdav::setMultiStatus {xmlText} {
	return [subst {<?xml version="1.0" encoding="utf-8"?>
<D:multistatus xmlns:D="DAV:">
$xmlText
</D:multistatus>}]
}

proc de::vogel::webdav::appendTextNodeChild {doc parent elementName data} {
	$parent appendChild [$doc createElement $elementName node]
	foreach {namespace tag} [split $elementName ":"] {break}
	switch -- $tag {
		creationdate {
			set data [clock format $data -format "%Y-%m-%dT%H:%M:%SZ" -gmt 1]
		}
		getlastmodified - lastaccessed {
			set data [clock format $data -format "%a, %d %b %Y %H:%M:%S %Z" -gmt 1]     
		}
		resourcetype - displayname - getcontentlength - getcontenttype - href - status {
			# do nothing: data is in correct form
		}
		default {
			error "de::vogel::webdav::appendTextNodeChild: Not supported elementName: '$elementName'!"
		}
	}
	$node appendChild [$doc createTextNode $data result]
	return $result
}


proc de::vogel::webdav::handler {prefix sock suffix} {
	global Httpd
	variable aConfig
	variable lPossibleMethods
	upvar #0 Httpd$sock globSock

	set method "$globSock(proto)"

	log "HANDLER BEGIN '$method' $globSock(uri)-----------------------------------------------"
	# some preconditions
	if {![info exists aConfig($prefix,namespace)]} {
		error "Namespace for '$prefix' doesn't exist!"
	}
	set namespace $aConfig($prefix,namespace)

	if {![info exists aConfig($prefix,initialized)]
		&& [set mInit [getMethod $prefix init]] != ""} {
		# first do cleanup
		if {[set mCleanup [getMethod $prefix cleanup]] != ""} {
			$mCleanup $prefix
		}
		$mInit $prefix
		set aConfig($prefix,initialized) 1
	}

	set doc ""
	set root ""
	set data ""
	if {[info exists globSock(mime,content-length)]} {
		set data [read $::env(HTTP_CHANNEL) $globSock(mime,content-length)]
	}

	set depth infinity
	if {[info exists globSock(mime,depth)]} {
		set depth $globSock(mime,depth)
	}
	set derivedMethod [getMethod $prefix $method]
	# TODO check for valid methods OPTIONS, PROPFIND, ???
	set result ""
	switch -- $method {
		PROPFIND {
			# analyse queried properties
			# dav-properties are:
			# creationdate, displayname, getcontentlanguage, getcontentlength, getcontenttype,
			# getetag, getlastmodified, lockdiscovery, resourcetype, source, supportedlock
			set lProperties {}
			set err ""
			if {![string length $data]
				|| [catch {set doc [dom parse -simple $data]} err]} {
				log "err: $err"
				set doc ""
			} else {
				set root [$doc documentElement]
				set prop [$root selectNodes *]
				if {[llength $prop] != 1} {
					error "Expected llength prop == 1 but [llength $prop]!"
				}
				
				log "Parsing Properties!!! '$data'"
				# oder /A:propfind/A:prop (wie findet man den namespace heras?)
				foreach node [$prop selectNodes *] {
					lappend lProperties [$node localName]
				}
			}

			log "lProperties: $lProperties"
			if {$derivedMethod != ""} {
				set result [$derivedMethod $globSock(uri) $depth $lProperties]
			} else {
				error "No such method '$method'!"
			}
		}
		OPTIONS {
			# if not defined by "mode" provide your own one
			if {$derivedMethod != ""} {
				set result [$derivedMethod $globSock(uri) $depth]
			} else {
				# o.k. try to determine yourself which methods are implemented

				# set lAllowedMethods {OPTIONS HEAD PROPPATCH POST}
				# check which methods are provided by module
				foreach m $lPossibleMethods {
					if {[getMethod $prefix $m] != ""} {
						lappend lAllowedMethods $m
					}
				}
				# [list MS-Author-Via: DAV] 
				# "Server: $Httpd(server)" 
				# [list Connection: Keep-Alive]
				# "Keep-Alive: timeout=15, max=98"
				# "Etag:"
				# [list DAV: 1,2]
				set lResultHeader [list \
									   {HTTP/1.1 200 OK} \
									   "Date: [HttpdDate [clock seconds]]" \
									   "Server: Tcl-DAV" \
									   [list DAV: 1] \
									   [list MS-Author-Via: DAV] \
									   [list Allow: [join $lAllowedMethods ,]] \
									   [list Content-Length: 0] \
									   [list Connection: close] \
									   [list Content-Type: httpd/unix-directory] \
									  ]
				Httpd_ReplaceHeader [Httpd_CurrentSocket] [join $lResultHeader \n]
				Httpd_ReturnData $sock text/html ""
			}
		}
		HEAD {
			if {$derivedMethod != ""} {
				set result [$derivedMethod $globSock(uri)]
			} else {
				error "$method must be implemented by namespace: '$namespace'"
			}
		}
		GET {
			if {$derivedMethod != ""} {
				set result [$derivedMethod $globSock(uri)]
			} else {
				error "$method must be implemented by namespace: '$namespace'"
			}
		}
		PUT {
			if {$derivedMethod != ""} {
				set result [$derivedMethod $globSock(uri) $data]
			} else {
				error "$method not supported: TODO!"
			}
		}
		POST {
			if {$derivedMethod != ""} {
				set result [$derivedMethod $globSock(uri) $data]
			} else {
				# error "$method not supported: TODO!"
				Httpd_ReturnData [Httpd_CurrentSocket] text/plain "Method POST is not allowed" 404
			}
		}
		DELETE {
			if {$derivedMethod != ""} {
				set result [$derivedMethod $globSock(uri)]
			} else {
				error ""
			}
		}
		MKCOL {
			if {$derivedMethod != ""} {
				set result [$derivedMethod $globSock(uri)]
			} else {
				Httpd_ReturnData [Httpd_CurrentSocket] text/plain "MKCOL is not allowed on resource '$namespace'." 403
			}
		}
		PROPPATCH {
			if {$derivedMethod != ""} {
			} else {
				error "TODO: $method Not allowed!"
			}
		}
		COPY - MOVE {
			if {$derivedMethod != ""} {
				# destination == http://localhost:8080/dav/....
				# overwrite-header: T (perform previous delete), F
				set overwrite ""
				if {[info exists globSock(mime,overwrite)]} {
					set overwrite $globSock(mime,overwrite)
				}
				set result [$derivedMethod $globSock(uri) $globSock(mime,destination) $overwrite $depth]
			} else {
				Httpd_ReturnData [Httpd_CurrentSocket] text/plain "COPY is not allowed in '$namespace' on '$globSock(uri)'." 403
			}
		}
		default {
			# uuups
			error "Unknown method: '$method'"
		}
	}
	if {$result != ""} {
		if {[llength $result] != 5} {
			error "result of '$derivedMethod' has wrong length should be 5 has [llength $result]"
		}
		foreach {returnCode addHeader mimeType returnData mode} $result {break}
		global Httpd_Errors
		global Httpd
		switch -- $mode {
			data {
				if {![info exists Httpd_Errors($returnCode)]} {
					error "No such return-code '$returnCode' (called from $derivedMethod) in Httpd_Errors: [array get Httpd_Errors]!"
				}
				# "Server: $Httpd(server)"
				# {Keep-Alive: timeout=15, max=98} 
				# {Connection: Keep-Alive}
				# "Etag: "
				set lHeader [list \
								 "HTTP/1.1 $returnCode $Httpd_Errors($returnCode)" \
								 "Date: [HttpdDate [clock seconds]]" \
								 "Server: Tcl-DAV" \
								 "Ms-Author-Via: DAV" \
								 "Connection: close" \
						 ]
				if {$mimeType != ""} {
					lappend lHeader "Content-Type: ${mimeType}; charset=\"utf-8\""
				} else {
					lappend lHeader "Content-Type: text/xml; charset=\"utf-8\""
				}
				if {[string first "transfer-encoding:" [string tolower $addHeader]] < 0} {
					lappend lHeader "Content-Length: [string length $returnData]"

				}
				log "lHeader: $lHeader addHeader: $addHeader"
				eval lappend lHeader $addHeader
				Httpd_ReplaceHeader [Httpd_CurrentSocket] [join $lHeader "\n"]
				Httpd_ReturnData [Httpd_CurrentSocket] $mimeType $returnData $returnCode
			}
			file {
				Httpd_AppendHeader [Httpd_CurrentSocket] $addHeader
				log "returnData: $returnData Mtype: [Mtype $returnData]"
				WebDav_ReturnFile [Httpd_CurrentSocket] [Mtype $returnData] $returnData
			}
			default {
				error "No such mode '$mode'. Must be {data, file}. Result from '$derivedMethod'"
			}
		}
	}
    # gather some useful parameters
    if {[namespace inscope :: "info procs ${namespace}::$method"] != ""} {
		# ::[set aConfig($prefix,namespace)]::$method $sock $globSock(uri) $depth
		# Httpd_ReturnData $sock text/html $html
	} else {
		# error "No such method implemented? $method"
	}
	log "HANDLER END '$method' -----------------------------------------------"
	catch {$doc delete}
}


unset -nocomplain de::vogel::webdav::aConfig
catch {namespace import de::vogel::webdav::webdav_resource}
de::vogel::webdav::readConfig "custom/webdav.conf"
