# ------------------------------------------------------------------------------
# WebDAV for Tclhttpd (webdav_file.tcl)
# 
# Support for WebDAV-Access to files in htdocs
# 
# for further instructions see webdav.tcl
# or 
# http://www.vogel-nest.de (go to "WebDAV")
# 
# HISTORY:
# 2004-07-29: Version 0.1a
#  * initial version
# ------------------------------------------------------------------------------
namespace eval de::vogel::webdav_file {

	variable lSupportedProperties
	set lSupportedProperties {
		getlastmodified 
		resourcetype
		getcontentlength
		displayname
		getcontenttype
		creationdate
	}
}

proc de::vogel::webdav_file::init {prefix} {
}

proc de::vogel::webdav_file::getProp {path propertiesList {forceDir 0}} {
	variable lSupportedProperties
	set path [string trimright $path /]
	set file [de::vogel::webdav::path2RealPath $path]
	de::vogel::webdav::log "webdav_file:getProp path: '$path' file: '$file'"
	if {[file isdirectory $file]} {
		append path /
	}
	set doc [dom createDocument D:response]
	set root [$doc documentElement]
	
	# trying to retrieve named properties
	if {[llength $propertiesList]} {
		$root setAttribute xmlns:g0 "DAV:"
	}

	if {[file isdirectory $file]} {
		set href [file tail $file]
	} else {
		set href [de::vogel::webdav::realPath2UrlPath $path]
	}
	de::vogel::webdav::appendTextNodeChild $doc $root D:href $href

	$root appendChild [$doc createElement D:propstat propstat]
	$propstat appendChild [$doc createElement D:prop prop]

    if {[file isdirectory $file] || $forceDir} {
		set resourceType "<D:collection/>"
		set contentLength ""
        set contentType ""
    } else {
		set resourceType ""
		set contentLength [file size $file]
        set contentType [Mtype $file]
        if {[string equal $contentType {*/*}]} {
            set contentType {text/plain}
        }
    }
	file stat $file stat
	de::vogel::webdav::appendTextNodeChild $doc $prop D:creationdate $stat(ctime)
	de::vogel::webdav::appendTextNodeChild $doc $prop D:getlastmodified [file mtime $file]
	# confuses IE
	# de::vogel::webdav::appendTextNodeChild $doc $prop D:lastaccessed [file atime $file]

	set node [de::vogel::webdav::appendTextNodeChild $doc $prop D:resourcetype $resourceType]
	if {$resourceType != ""} {
		$node disableOutputEscaping 1
	}

	de::vogel::webdav::appendTextNodeChild $doc $prop D:getcontentlength $contentLength
	de::vogel::webdav::appendTextNodeChild $doc $prop D:displayname [file tail $file]
	de::vogel::webdav::appendTextNodeChild $doc $prop D:getcontenttype $contentType
	de::vogel::webdav::appendTextNodeChild $doc $propstat D:status {HTTP/1.1 200 OK}

	if {[llength $propertiesList]} {
		# determine not already provided properties
		# and set them to 404 (not available)
		set lNotAlreadyProvided [::struct::set::Sdifference $propertiesList $lSupportedProperties]
		if {[llength $lNotAlreadyProvided]} {
			$root appendChild [$doc createElement D:propstat propstat]
			$propstat appendChild [$doc createElement D:prop prop]
			foreach property $lNotAlreadyProvided {
				$prop appendChild [$doc createElement D:$property]
			}
			$propstat appendChild [$doc createElement D:status node]
			$node appendChild [$doc createTextNode {HTTP/1.1 404 Not Found}]
		}
	}

    set xml [$root asXML]
    $doc delete

    return $xml
}

proc de::vogel::webdav_file::PROPFIND {path depth propertiesList} {
	global Httpd
	global Config

	de::vogel::webdav::log "webdav_file::PROPFIND path: '$path' depth: '$depth' propertiesList: '$propertiesList'"
	set result {}

	set realPath [de::vogel::webdav::path2RealPath $path]
	if {![file exists $realPath]} {
		# no file or collection - no fun
		return [list 404 "" text/plain "File: '$path' not found" data]
	}

	if {$depth != "0"} {
		set entries [glob -directory $Config(docRoot) -nocomplain [file join $path *]]
	} else {
		set entries $path
	}
	if {[string length $entries] == 0} {
		set entries $path
	}
	de::vogel::webdav::log "entries: $entries"
	foreach entry $entries {
		append result [de::vogel::webdav_file::getProp $entry $propertiesList]
	}
	set result [de::vogel::webdav::setMultiStatus $result]
	return [list 207 "" text/xml $result data]
}

proc de::vogel::webdav_file::HEAD {file} {
	set realPath [de::vogel::webdav::path2RealPath $file]
	de::vogel::webdav::log "webdav_file::HEAD: $file $realPath"
	if {![file exists $realPath]} {
		return [list 404 "" text/plain "File: '$file' not found" data]
	} else {
		return [list 200 "" "" $realPath data]
	}
}

proc de::vogel::webdav_file::GET {file} {
	global Config
	set realPath [de::vogel::webdav::path2RealPath $file]
	de::vogel::webdav::log "webdav_file::GET '$file' realPath: $realPath"
	if {![file exists $realPath]} {
		return [list 404 "" text/plain "File: '$file' not found" data]
	} elseif {[file isdirectory $realPath]} {
		set dirListing [glob -tails -directory [file join $Config(docRoot) [string trimleft $file /]] -nocomplain *]
		return [list 200 "" text/plain $dirListing data]
	} else {
		return [list "" "" "" $realPath file]
	}
}

proc de::vogel::webdav_file::PUT {file data} {
	set realPath [de::vogel::webdav::path2RealPath $file]
	if {[catch {
		set fh [open $realPath w+] 
		fconfigure $fh -translation binary
		puts $fh $data
		close $fh
	} msg]} {
		return [list 409 "" text/plain "Error creating file: '$file': $msg" data]
	} else {
		return [list 201 "" text/plain "File: '$file' created." data]
	}
}

proc de::vogel::webdav_file::MKCOL {path} {
	set realDir [de::vogel::webdav::path2RealPath $path]
	if {[catch {file mkdir $realDir} msg]} {
		return [list 403 "" text/plain "Error creating collection '$path': $msg" data]
	} else {
		return [list 201 "" text/plain "Collection '$path' created." data]
	}
}

proc de::vogel::webdav_file::DELETE {path} {
	set realDir [de::vogel::webdav::path2RealPath $path]
	if {[catch {file delete $realDir} msg]} {
		# TODO must be MULTISTATUS
		return [list 207 "" text/plain "Error creating collection '$path': $msg" data]
	} else {
		# No Content -- successful deletion
		return [list 204 "" text/plain "'$path' deleted." data]
	}
}

proc de::vogel::webdav_file::COPY {path destination overwrite depth} {
	de::vogel::webdav::log "called webdav_file::COPY: '$path' '$destination' '$depth'"
	set realPath [de::vogel::webdav::path2RealPath $path]
	array set aDest [uri::split $destination]
	set realDestpath [de::vogel::webdav::path2RealPath $aDest(path)]
	if {![catch {
		file copy $realPath $realDestpath
	} err]} {
		return [list 200 "" "" $realPath data]
	} else {
		# error-message
		de::vogel::webdav::log "err: $err"
	}
}

proc de::vogel::webdav_file::MOVE {path destination overwrite depth} {
	de::vogel::webdav::log "webdav_file::MOVE: path: '$path' destination: '$destination' overwrite: '$overwrite' depth: '$depth'"
	set realPath [de::vogel::webdav::path2RealPath $path]
	array set aDest [uri::split $destination]
	set realDestpath [de::vogel::webdav::path2RealPath $aDest(path)]
	if {![catch {
		set fileAlreadyExists [file exists $realDestpath]
		file rename $realPath $realDestpath
	} err]} {
		if {$fileAlreadyExists} {
			# No Content
			return [list 204 "" "" $realDestpath data]
		} else {
			# Created
			return [list 201 "" "" $realDestpath data]
		}
	} else {
		# error-message
		de::vogel::webdav::log "webdav_file:MOVE err: $err"
	}
}